#ifdef _WINDOWS
#ifndef WINDOWS
#define WINDOWS
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>

#include "ac_plugin.h"
#include "cmshgroup.h"
#include "cmshmat.h"
#include "cmshtex.h"

//-----------------------------------------------------------------------------------------
#define MSH_LINE_LEN 1024
void getmshdata( char *dataline, FILE *fp, int *line )
{
   dataline = fgets( dataline, MSH_LINE_LEN - 1, fp );
   ( *line ) += 1;
   if( dataline == NULL )
   {
      throw "msh : end of file encountered";
   }
   // Chop off newlines
   int slen = ( int )strlen( dataline );
   if( slen == 0 )
   {
      return;
   }
   char *p = &( dataline[slen - 1] );
   while( *p == ( char )10 || *p == ( char )13 )
   {
      *p = '\0';
      if( p == dataline )
      {
         break;
      }
      --p;
   }
}
//-----------------------------------------------------------------------------------------
#define MSHTYPE_NA         -1
#define MSHTYPE_GROUPS      0
#define MSHTYPE_MATERIALS   1
#define MSHTYPE_TEXTURES    2
#define MSHTYPE_G_MATERIAL  4
#define MSHTYPE_G_TEXTURE   5
#define MSHTYPE_G_TEXWRAP   6
#define MSHTYPE_G_NONORMAL  7
#define MSHTYPE_G_FLAG      8
#define MSHTYPE_G_GEOM      9
int getmshltype( char *token, char *dataline )
{
   // TODO: Find out whether these tags are case-sensitive
   if( sscanf( dataline, "%s", token ) != 1 )
   {
      token[0] = '\0';
      return MSHTYPE_NA;
   }
   else if( strcmp( token, "GROUPS" ) == 0 ) return MSHTYPE_GROUPS;
   else if( strcmp( token, "MATERIALS" ) == 0 ) return MSHTYPE_MATERIALS;
   else if( strcmp( token, "TEXTURES" ) == 0 ) return MSHTYPE_TEXTURES;
   else if( strcmp( token, "MATERIAL" ) == 0 ) return MSHTYPE_G_MATERIAL;
   else if( strcmp( token, "TEXTURE" ) == 0 ) return MSHTYPE_G_TEXTURE;
   else if( strcmp( token, "TEXWRAP" ) == 0 ) return MSHTYPE_G_TEXWRAP;
   else if( strcmp( token, "NONORMAL" ) == 0 ) return MSHTYPE_G_NONORMAL;
   else if( strcmp( token, "FLAG" ) == 0 ) return MSHTYPE_G_FLAG;
   else if( strcmp( token, "GEOM" ) == 0 ) return MSHTYPE_G_GEOM;
   else return MSHTYPE_NA;
}
//-----------------------------------------------------------------------------------------
static ACObject *do_msh_load( char *filename )
{

   FILE *fp = fopen( filename, "r" );
   if( fp == NULL) { 
      message_dialog( "error opening file \"%s\"", filename );
      return NULL;
   }

   // Unbackslash the path
#ifdef WINDOWS
   ac_unbackslash_string( filename );
#endif
   // Find the last directory delimiter in filename, replace the next char with a null-terminator.
   // This will become our search path for textures, since Orbiter doesn't add a path to textures.
   char texpath[2048];
   strncpy( texpath, filename, 2047 );
   char *pstr = strrchr( texpath, '/' );
   if( pstr == NULL )
   {
      texpath[0] = '\0';
   }
   else
   {
      ++pstr;
      *pstr = '\0';
   }

   char dataline[MSH_LINE_LEN];
   int line = 0;
   char token[MSH_LINE_LEN];
   char token2[MSH_LINE_LEN];
   char read_activity[1024];
   int numgroups = 0;
   int nummats = 0;
   int numtexs = 0;
   CMshGroup *msharr = NULL;
   CMshMat *matarr = NULL;
   CMshTex *texarr = NULL;
   int i1, i2;
   float f1, f2, f3;

   bool success = false;

   try
   {

      // Grab header
      strcpy( read_activity, "reading MSH header" );
      getmshdata( dataline,  fp, &line );
      if( strncmp( "MSHX1", dataline, 5 ) != 0 )
      {
         throw "invalid MSH header";
      }

      // Get # of groups
      strcpy( read_activity, "reading number of groups" );
      int type = MSHTYPE_NA;
      while( type != MSHTYPE_GROUPS )
      {
         getmshdata( dataline, fp, &line );
         if( getmshltype( token, dataline ) == MSHTYPE_GROUPS )
         {
            if( sscanf( dataline, "%s %d", token, &numgroups ) != 2 )
            {
               throw "error reading number of groups";
            }
            break;
         }
      }

      // Allocate an array of groups
      msharr = new CMshGroup[numgroups];

      // Read each group
      for( int ix = 0; ix < numgroups; ++ix )
      {

         sprintf( read_activity, "reading group %d", ix + 1 );
         CMshGroup *msh = &( msharr[ix] );

         // Read until we hit a GEOM group
         type = MSHTYPE_NA;
         while( type != MSHTYPE_G_GEOM )
         {
            getmshdata( dataline, fp, &line );
            type = getmshltype( token, dataline );
            switch( type )
            {
            case MSHTYPE_G_MATERIAL:
               {
                  sscanf( dataline, "%s %d", token, &( msh->mMatIdx ));
                  msh->mMatIdx -= 1;
                  break;
               }
            case MSHTYPE_G_TEXTURE:
               {
                  sscanf( dataline, "%s %d", token, &( msh->mTexIdx ));
                  msh->mTexIdx -= 1;
                  break;
               }
            case MSHTYPE_G_TEXWRAP:
               {
                  sscanf( dataline, "%s %s", token, token2 );
                  msh->mTexWrap[0] = token2[0];
                  msh->mTexWrap[1] = token2[1];
                  msh->mTexWrap[2] = '\0';
                  break;
               }
            case MSHTYPE_G_NONORMAL:
               {
                  msh->mNoNormal = true;
                  break;
               }
            case MSHTYPE_G_FLAG:
               {
                  sscanf( dataline, "%s %d", token, &( msh->mFlag ));
                  break;
               }
            case MSHTYPE_G_GEOM:
               {
                  sscanf( dataline, "%s %d %d", token, &i1, &i2 );
                  msh->AllocateGeom( i1, i2 );
                  for( int iy = 0; iy < i1; ++iy )
                  {
                     // TODO: Find out whether we can have tex coords w/o normals
                     getmshdata( dataline, fp, &line );
                     sscanf( dataline, "%f %f %f %f %f %f %f %f",
                          &( msh->mVerts[iy][0] ),
                          &( msh->mVerts[iy][1] ),
                          &( msh->mVerts[iy][2] ),
                             &f1, &f2, &f3,
                          &( msh->mVerts[iy][3] ),
                          &( msh->mVerts[iy][4] ));
                  }
                  for( int iy = 0; iy < i2; ++iy )
                  {
                     getmshdata( dataline, fp, &line );
                     sscanf( dataline, "%d %d %d",
                          &( msh->mTris[iy][0] ),
                          &( msh->mTris[iy][1] ),
                          &( msh->mTris[iy][2] ));
                  }
                  break;
               }
            }
         }

      }

      // Get materials
      strcpy( read_activity, "reading number of materials" );
      type = MSHTYPE_NA;
      while( type != MSHTYPE_MATERIALS )
      {
         getmshdata( dataline, fp, &line );
         if( getmshltype( token, dataline ) == MSHTYPE_MATERIALS )
         {
            if( sscanf( dataline, "%s %d", token, &nummats ) != 2 )
            {
               throw "error reading number of materials";
            }
            break;
         }
      }

      // Allocate an array of groups
      matarr = new CMshMat[nummats];

      // Read each material
      for( int ix = 0; ix < nummats; ++ix )
      {

         sprintf( read_activity, "reading material %d", ix + 1 );
         CMshMat *msh = &( matarr[ix] );

         // Read until we hit a GEOM group
         type = MSHTYPE_NA;
         while( type != MSHTYPE_G_MATERIAL )
         {
            getmshdata( dataline, fp, &line );
            type = getmshltype( token, dataline );
            switch( type )
            {
            case MSHTYPE_G_MATERIAL:
               {
                  if( strlen( dataline ) > 8 )
                  {
                     strcpy( token, dataline + 9 );
                     msh->SetName( token );
                  }
                  getmshdata( dataline, fp, &line );
                  sscanf( dataline, "%f %f %f %f",
                     &( msh->mDiffuse[0] ),
                     &( msh->mDiffuse[1] ),
                     &( msh->mDiffuse[2] ),
                     &( msh->mDiffuse[3] ));
                  getmshdata( dataline, fp, &line );
                  sscanf( dataline, "%f %f %f %f",
                     &( msh->mAmbient[0] ),
                     &( msh->mAmbient[1] ),
                     &( msh->mAmbient[2] ),
                     &( msh->mAmbient[3] ));
                  getmshdata( dataline, fp, &line );
                  sscanf( dataline, "%f %f %f %f %f",
                     &( msh->mSpecular[0] ),
                     &( msh->mSpecular[1] ),
                     &( msh->mSpecular[2] ),
                     &( msh->mSpecular[3] ),
                     &( msh->mSpecular[4] ));
                  getmshdata( dataline, fp, &line );
                  sscanf( dataline, "%f %f %f %f",
                     &( msh->mEmissive[0] ),
                     &( msh->mEmissive[1] ),
                     &( msh->mEmissive[2] ),
                     &( msh->mEmissive[3] ));
                  break;
               }
            }
         }

      }

      // Get textures
      strcpy( read_activity, "reading number of textures" );
      type = MSHTYPE_NA;
      while( type != MSHTYPE_TEXTURES )
      {
         getmshdata( dataline, fp, &line );
         if( getmshltype( token, dataline ) == MSHTYPE_TEXTURES )
         {
            if( sscanf( dataline, "%s %d", token, &numtexs ) != 2 )
            {
               throw "error reading number of textures";
            }
            break;
         }
      }

      // Allocate an array of groups
      texarr = new CMshTex[numtexs];

      // Read each texture
      for( int ix = 0; ix < numtexs; ++ix )
      {
         CMshTex *msh = &( texarr[ix] );
         getmshdata( dataline, fp, &line );
         msh->SetName( dataline );
      }

      success = true;

   }
   catch( char *err )
   {
      printf( "MSH error:\n%s at line %d while %s\nLast known dataline was %s\n", err, line, read_activity, dataline );
      message_dialog( "msh : %s at line %d while %s", err, line, read_activity );
   }

   // Close input file
   fclose( fp );

   ACObject *retobj = NULL;

   // Add items if we finished a successful read
   if( success )
   {

      // Add the materials
      int *matidx = NULL;
      if( nummats > 0 )
      {
         matidx = new int[nummats];
      }
      for( int ix = 0; ix < nummats; ++ix )
      {
         CMshMat *mat = &( matarr[ix] );
         ACMaterialTemplate *mt = new ACMaterialTemplate();
         mt->ambient.r = mat->mAmbient[0];
         mt->ambient.g = mat->mAmbient[1];
         mt->ambient.b = mat->mAmbient[2];
         mt->rgb.r = mat->mDiffuse[0];
         mt->rgb.g = mat->mDiffuse[1];
         mt->rgb.b = mat->mDiffuse[2];
         mt->specular.r = mat->mSpecular[0];
         mt->specular.g = mat->mSpecular[1];
         mt->specular.b = mat->mSpecular[2];
         mt->emissive.r = mat->mEmissive[0];
         mt->emissive.g = mat->mEmissive[1];
         mt->emissive.b = mat->mEmissive[2];
         mt->shininess = mat->mSpecular[4];
         mt->transparency = min( 1.0f, max( 0.0f, 1.0f - mat->mDiffuse[3] ));
         // Add the material, get the index
         matidx[ix] = ac_palette_get_new_material_index( mt );
         ACMaterial *m = ac_palette_get_material( matidx[ix] );
         ac_material_set_name( m, mat->mName );
      }

      // Add the textures
      char texfilename[2650];
      for( int ix = 0; ix < numtexs; ++ix )
      {
         CMshTex *tex = &( texarr[ix] );
         strcpy( texfilename, texpath );
         strcat( texfilename, tex->mName );
         if( _access( texfilename, 0 ) == 0 )
         {
            tex->mIndex = add_new_texture_opt( texfilename, tex->mName );
         }
         else
         {
            printf( "Could not find texture \"%s\"\n", texfilename );
         }
      }

      // Add each object
      List *objlist = NULL;
      for( int ix = 0; ix < numgroups; ++ix )
      {
         CMshGroup *grp = &( msharr[ix] );
         // Add verts
         if( grp->mNumVerts > 0 )
         {
            ACObject *obj = ac_new_object();
            if( grp->mName == NULL )
            {
               object_set_name( obj, "msh_geom" );
            }
            else
            {
               object_set_name( obj, grp->mName );
            }
            Vertex **varr = new Vertex*[grp->mNumVerts];
            for( int iy = 0; iy < grp->mNumVerts; ++iy )
            {
               varr[iy] = new_vertex_set( -( grp->mVerts[iy][0] ), grp->mVerts[iy][1], grp->mVerts[iy][2] );
               object_add_vertex_head( obj, varr[iy] );
            }
            for( int iy = 0; iy < grp->mNumTris; ++iy )
            {
               Surface *surf = new_surface();
               object_add_surface_head( obj, surf );
               int i1 = grp->mTris[iy][0];
               int i2 = grp->mTris[iy][1];
               int i3 = grp->mTris[iy][2];
               surface_add_vertex( surf, varr[i3], ( grp->mVerts[i3][3] ), -( grp->mVerts[i3][4] ));
               surface_add_vertex( surf, varr[i2], ( grp->mVerts[i2][3] ), -( grp->mVerts[i2][4] ));
               surface_add_vertex( surf, varr[i1], ( grp->mVerts[i1][3] ), -( grp->mVerts[i1][4] ));
               surface_set_shading( surf, TRUE );
               surface_set_twosided( surf, false );
               if( grp->mMatIdx > -1 )
               {
                  surface_set_col( surf, matidx[grp->mMatIdx] );
               }
            }
            delete[] varr;
            // Set the texture
            if( grp->mTexIdx > -1 && texarr[grp->mTexIdx].mIndex > -1 )
            {
               object_texture_set( obj, texarr[grp->mTexIdx].mIndex );
            }
            ac_object_set_crease_angle( obj, 180.0f );
            object_calc_normals_force( obj );
            list_add_item_head( &objlist, obj );
         }
      }

      delete[] matidx;

      retobj = ac_new_world();
      object_add_children( retobj, objlist );
      ac_selection_select_objectlist( objlist );
      list_free( &objlist );
      fit_selected_all();

   }

   // Create objects

   // Cleanup objects
   if( msharr != NULL )
   {
      delete[] msharr;
   }
   if( matarr != NULL )
   {
      delete[] matarr;
   }
   if( texarr != NULL )
   {
      delete[] texarr;
   }

   return retobj;

}
#undef MSHTYPE_NA
#undef MSHTYPE_GROUPS
#undef MSHTYPE_MATERIALS
#undef MSHTYPE_TEXTURES
#undef MSHTYPE_G_MATERIAL
#undef MSHTYPE_G_TEXTURE
#undef MSHTYPE_G_TEXWRAP
#undef MSHTYPE_G_NONORMAL
#undef MSHTYPE_G_FLAG
#undef MSHTYPE_G_GEOM
#undef MSHTYPE_M_MATERIAL
#undef MSH_LINE_LEN
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *data )
{

   static bool firstTime = true;
   if( firstTime )
   {

      double ver = ac_get_version_number();

      firstTime = false;

      if (ver < 3.7)
      {
         printf( "ERROR - msh importer requires AC3D 3.7 or later" );
         return -1;
      }

      ac_register_file_importer( "MSH", ".msh", "Orbiter MSH files", do_msh_load, "plugin, version 1.0, by Dennis Hawthorne" );

   }

   return 0;

}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Orbiter MSH Import - Version 1.0 - by Dennis Hawthorne" );
}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Orbiter MSH importer." );
}
//-----------------------------------------------------------------------------------------
